import unittest

from neo3.wallet import account, scrypt_parameters as scrypt

account_list = [
    {
        "address": "NRaKbRA5JAEJtfUgJJZzmeDnKvP3pJwKp1",
        "encrypted_key": "6PYKuriAL7pFeVTr3tKksbD1SpKUP7K82vjGuskZ5zpo9EWDhLRW6GcnyL",
        "password": "city of zion",
        "private_key": "58124574dfcca1a7a958775f6ea94e3d6c392ec3ba125b5bc591dd5e14f05e52",
        "script_hash": "18f13748e08d53c9a164227e1a3e8d8d9e78193e",
        "wif_key": "KzAuju4yBqBhmUzYpfEEppPW8jfxALTsdsUR8hLPv9R3PBD97CUv",
    },
    {
        "address": "NgPptMp2tcjnXuYbUrTozvwvLExGKk5jXc",
        "encrypted_key": "6PYMEujkLZiJrQ5AK9W4z1BtYZT2U27ZVKrjbEFt8zZh5CJANZdEx21Fyx",
        "password": "123",
        "private_key": "2032b737522d22e2b6faf30555faa91d95c5aa5113c18f218f45815b6934c558",
        "script_hash": "cfa9032d65b3d0fc1df3956a4ef01666f23ba7e0",
        "wif_key": "KxJJLmU1Nv7igx3RFM4siSvio7wasF3ZzMzi7SrJ1s78QDQeEtjs",
        "scrypt": {"n": 2, "r": 8, "p": 8},
    },
    {
        "address": "NZMHRJMPbyJJwtXpvS2mYAWcWp4qmZZFx8",
        "encrypted_key": "6PYL44vbRemjfwCJ8qprKKJJiuzcopnJhghPoMLRVJLpymDwm2BNj9v7fq",
        "password": "neo",
        "private_key": "4c5182d9041f416bee1a6adac6a03f3e0319a83e75e78e6ff739304095791f19",
        "script_hash": "0df27baba6baeeb6834bea0d6c2a78183b416393",
        "wif_key": "Kyn4fA6czAhktoAM9YXKv3m7jtt47AuQxCXqSusnBmj3GsZUZQ6M",
        "scrypt": {"n": 2, "r": 8, "p": 8},
    },
]


class AccountCreationTestCase(unittest.TestCase):
    def test_new_account(self):
        for testcase in account_list[1:]:
            scrypt_params = testcase.get("scrypt", None)
            if scrypt_params is not None:
                scrypt_params = scrypt.ScryptParameters.from_json(scrypt_params)
            acc = account.Account(scrypt_parameters=scrypt_params)
            self.assertIsNotNone(acc)
            self.assertIsNotNone(acc.address)
            self.assertIsNotNone(acc.private_key)
            self.assertIsNotNone(acc.public_key)

    def test_new_account_from_private_key(self):
        for testcase in account_list:
            scrypt_params = testcase.get("scrypt", None)
            if scrypt_params is not None:
                scrypt_params = scrypt.ScryptParameters.from_json(scrypt_params)
            acc = account.Account.from_private_key(
                bytes.fromhex(testcase["private_key"]), scrypt_params
            )
            self.assertEqual(testcase["address"], acc.address)
            self.assertEqual(
                testcase["encrypted_key"].encode("utf-8"),
                account.Account.private_key_to_nep2(
                    bytes.fromhex(testcase["private_key"]),
                    testcase["password"],
                    scrypt_params,
                ),
            )
            self.assertEqual(testcase["script_hash"], str(acc.script_hash))
            self.assertIsNotNone(acc.public_key)

    def test_new_account_from_encrypted_key(self):
        for testcase in account_list[1:]:
            scrypt_params = testcase.get("scrypt", None)
            if scrypt_params is not None:
                scrypt_params = scrypt.ScryptParameters.from_json(scrypt_params)

            acc = account.Account.from_encrypted_key(
                testcase["encrypted_key"], testcase["password"], scrypt_params
            )
            self.assertEqual(testcase["address"], acc.address)
            self.assertEqual(
                testcase["encrypted_key"].encode("utf-8"),
                account.Account.private_key_to_nep2(
                    bytes.fromhex(testcase["private_key"]),
                    testcase["password"],
                    scrypt_params,
                ),
            )
            self.assertEqual(testcase["script_hash"], str(acc.script_hash))
            self.assertIsNotNone(acc.public_key)

    def test_new_watch_only_account(self):
        from neo3.core.types import UInt160

        for testcase in account_list[1:]:
            acc = account.Account.watch_only(
                UInt160.from_string(testcase["script_hash"])
            )
            self.assertEqual(testcase["address"], acc.address)
            self.assertIsNone(acc.private_key)
            self.assertEqual(testcase["script_hash"], str(acc.script_hash))
            self.assertIsNone(acc.public_key)
            self.assertTrue(acc.is_watchonly)

    def test_new_watch_only_account_from_address(self):
        for testcase in account_list[1:]:
            acc = account.Account.watch_only_from_address(testcase["address"])
            self.assertEqual(testcase["address"], acc.address)
            self.assertIsNone(acc.private_key)
            self.assertEqual(testcase["script_hash"], str(acc.script_hash))
            self.assertIsNone(acc.public_key)

    def test_new_account_from_wif(self):
        for testcase in account_list[:1]:
            scrypt_params = testcase.get("scrypt", None)
            if scrypt_params is not None:
                scrypt_params = scrypt.ScryptParameters.from_json(scrypt_params)

            acc = account.Account.from_wif(testcase["wif_key"], scrypt_params)
            self.assertEqual(testcase["address"], acc.address)
            self.assertEqual(
                testcase["encrypted_key"].encode("utf-8"),
                account.Account.private_key_to_nep2(
                    bytes.fromhex(testcase["private_key"]),
                    testcase["password"],
                    scrypt_params,
                ),
            )
            self.assertEqual(testcase["script_hash"], str(acc.script_hash))
            self.assertIsNotNone(acc.public_key)

    def test_new_account_wrong_password(self):
        for testcase in account_list:
            with self.assertRaises(ValueError) as context:
                account.Account.from_encrypted_key(
                    testcase["encrypted_key"], "wrong password"
                )
            self.assertIn("Wrong passphrase", str(context.exception))

    def test_to_wif(self):
        wif = "L5kx9QRKG9dwzSJF72pgps1d2scJZjnECWoKuUGVsz2D1WRBEaJ7"
        acc = account.Account.from_wif(wif)
        self.assertEqual(wif, account.Account.private_key_to_wif(acc.private_key))



"""
This module holds helper functions for data that has been serialized using the StdLib native contract
"""

from neo3 import vm
from typing import NamedTuple, cast, Any
from neo3.core import serialization, types


class PlaceHolder(NamedTuple):
    type: vm.StackItemType
    count: int  # type: ignore


def binary_deserialize(data: bytes):
    """
    Deserialize data that has been serialized using StdLib.serialize()

    This is the equivalent of the StdLib.deserialize()
    https://github.com/neo-project/neo/blob/29fab8d3f8f21046a95232b29053c08f9d81f0e3/src/Neo/SmartContract/Native/StdLib.cs#L39
    and can be used to deserialize data from smart contract storage that was serialized when stored.
    """
    # https://github.com/neo-project/neo-vm/blob/859417ad8ff25c2e4a432b6b5b628149875b3eb9/src/Neo.VM/ExecutionEngineLimits.cs#L39
    max_size = 0xFFFF * 2
    if len(data) == 0:
        raise ValueError("Nothing to deserialize")

    deserialized: list[Any | PlaceHolder] = []
    to_deserialize = 1
    with serialization.BinaryReader(data) as reader:
        while not to_deserialize == 0:
            to_deserialize -= 1
            item_type = vm.StackItemType(reader.read_byte()[0])
            if item_type == vm.StackItemType.ANY:
                deserialized.append(None)
            elif item_type == vm.StackItemType.BOOLEAN:
                deserialized.append(reader.read_bool())
            elif item_type == vm.StackItemType.INTEGER:
                # https://github.com/neo-project/neo-vm/blob/859417ad8ff25c2e4a432b6b5b628149875b3eb9/src/Neo.VM/Types/Integer.cs#L27
                deserialized.append(int(types.BigInteger(reader.read_var_bytes(32))))
            elif item_type in [vm.StackItemType.BYTESTRING, vm.StackItemType.BUFFER]:
                deserialized.append(reader.read_var_bytes(len(data)))
            elif item_type in [vm.StackItemType.ARRAY, vm.StackItemType.STRUCT]:
                count = reader.read_var_int(max_size)
                deserialized.append(PlaceHolder(item_type, count))
                to_deserialize += count
            elif item_type == vm.StackItemType.MAP:
                count = reader.read_var_int(max_size)
                deserialized.append(PlaceHolder(item_type, count))
                to_deserialize += count * 2
            else:
                raise ValueError("unreachable")

    temp: list = []
    while len(deserialized) > 0:
        item = deserialized.pop()
        if type(item) == PlaceHolder:
            item = cast(PlaceHolder, item)
            if item.type == vm.StackItemType.ARRAY:
                array = []
                for _ in range(0, item.count):
                    array.append(temp.pop())
                temp.append(array)
            elif item.type == vm.StackItemType.STRUCT:
                struct = []
                for _ in range(0, item.count):
                    struct.append(temp.pop())
                temp.append(struct)
            elif item.type == vm.StackItemType.MAP:
                m = dict()
                for _ in range(0, item.count):
                    k = temp.pop()
                    v = temp.pop()
                    m[k] = v
                temp.append(m)
        else:
            temp.append(item)
    return temp.pop()



import unittest
from typing import Optional, Any
from aioresponses import aioresponses
from neo3.core import types
from neo3.api.wrappers import _check_address_and_convert, ChainFacade

JSON = Any


class WrapperUtilsTest(unittest.TestCase):
    def test_check_address_and_convert(self):
        hash_in = types.UInt160.from_string(
            "0x7e9237a93f64407141a5b86c760200c66c81e2ec"
        )
        self.assertIsInstance(_check_address_and_convert(hash_in), types.UInt160)

        with self.assertRaises(ValueError) as context:
            _check_address_and_convert(object())
        self.assertEqual(
            "Input is of type <class 'object'> expected UInt160 or NeoAddress(str)",
            str(context.exception),
        )

        invalid_address = "NgNJsBfhcoJSm6MVMpMeGLqEK5mSQXuJTt"
        with self.assertRaises(ValueError) as context:
            _check_address_and_convert(invalid_address)
        self.assertEqual("Invalid checksum", str(context.exception))

        valid_address = "NgNJsBfhcoJSm6MVMpMeGLqEK5mSQXuJTq"
        self.assertIsInstance(_check_address_and_convert(valid_address), types.UInt160)


class TestChainFacade(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        # CAREFULL THIS PATCHES ALL aiohttp CALLS!
        self.helper = aioresponses()
        self.helper.start()

    async def asyncTearDown(self) -> None:
        self.helper.stop()

    def mock_response(
        self, payload: Optional[JSON] = None, exc: Optional[Exception] = None
    ):
        """
        Either payload or exc should be provided
        """
        if payload is not None and exc is not None:
            raise ValueError("Arguments are mutual exclusive")

        if payload is not None:
            json = {"jsonrpc": "2.0", "id": 1, "result": payload}
            self.helper.post("localhost", payload=json)
        else:
            self.helper.post("localhost", exception=exc)

    async def test_receipt_retry_delay_and_timeout(self):
        user_agent = "/Neo:3.0.3/"
        get_version_captured = {
            "tcpport": 10333,
            "wsport": 10334,
            "nonce": 1930156121,
            "useragent": user_agent,
            "rpc": {"maxiteratorresultitems": 100, "sessionenabled": True},
            "protocol": {
                "addressversion": 53,
                "network": 860833102,
                "validatorscount": 7,
                "msperblock": 15000,
                "maxtraceableblocks": 2102400,
                "maxvaliduntilblockincrement": 5760,
                "maxtransactionsperblock": 512,
                "memorypoolmaxtransactions": 50000,
                "initialgasdistribution": 5200000000000000,
                "hardforks": [
                    {"name": "Aspidochelone", "blockheight": 1730000},
                    {"name": "Basilisk", "blockheight": 4120000},
                    {"name": "Cockatrice", "blockheight": 5450000},
                    {"name": "Domovoi", "blockheight": 5570000},
                ],
            },
        }
        self.mock_response(get_version_captured)
        facade = ChainFacade("localhost")
        delay, timeout = await facade._get_receipt_time_values()
        self.assertEqual(3.0, delay)
        self.assertEqual(33.0, timeout)

        self.mock_response(get_version_captured)
        facade = ChainFacade("localhost", receipt_timeout=1)
        delay, timeout = await facade._get_receipt_time_values()
        self.assertEqual(3.0, delay)
        self.assertEqual(1.0, timeout)

        self.mock_response(get_version_captured)
        facade = ChainFacade("localhost", receipt_retry_delay=5)
        delay, timeout = await facade._get_receipt_time_values()
        self.assertEqual(5.0, delay)
        self.assertEqual(35.0, timeout)

        self.mock_response(get_version_captured)
        facade = ChainFacade("localhost", receipt_retry_delay=5, receipt_timeout=1)
        delay, timeout = await facade._get_receipt_time_values()
        self.assertEqual(5.0, delay)
        self.assertEqual(1.0, timeout)



"""
Block payload and related classes.
"""

from __future__ import annotations
import hashlib
from neo3.core import Size as s, serialization, types, utils, cryptography as crypto
from neo3.network.payloads import verification, transaction, inventory
from bitarray import bitarray  # type: ignore
from collections.abc import Sequence
from typing import Optional


class Header(verification.IVerifiable):
    """
    A `Block` header only object.

    Does not contain any consensus data or transactions.

    See also:
        :class:`~neo3.network.payloads.block.TrimmedBlock`
    """

    def __init__(
        self,
        version: int,
        prev_hash: types.UInt256,
        timestamp: int,
        nonce: int,
        index: int,
        primary_index: int,
        next_consensus: types.UInt160,
        witness: verification.Witness,
        merkle_root: Optional[types.UInt256] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.version = version
        self.prev_hash = prev_hash
        self.merkle_root = merkle_root if merkle_root else types.UInt256.zero()
        self.timestamp = timestamp
        self.nonce = nonce
        self.index = index
        self.primary_index = primary_index
        self.next_consensus = next_consensus
        self.witness = witness

    def __len__(self):
        return (
            s.uint32
            + s.uint256
            + s.uint256
            + s.uint64
            + s.uint64
            + s.uint32
            + s.uint8
            + s.uint160
            + 1
            + len(self.witness)
        )

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    def hash(self) -> types.UInt256:
        """
        Get a unique identifier based on the unsigned data portion of the object.
        """
        with serialization.BinaryWriter() as bw:
            self.serialize_unsigned(bw)
            data_to_hash = bytearray(bw._stream.getvalue())
            data = hashlib.sha256(data_to_hash).digest()
            return types.UInt256(data=data)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        self.serialize_unsigned(writer)
        writer.write_serializable_list([self.witness])

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned data portion.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.version)
        writer.write_serializable(self.prev_hash)
        writer.write_serializable(self.merkle_root)
        writer.write_uint64(self.timestamp)
        writer.write_uint64(self.nonce)
        writer.write_uint32(self.index)
        writer.write_uint8(self.primary_index)
        writer.write_serializable(self.next_consensus)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the check byte does not equal.
        """
        self.deserialize_unsigned(reader)
        witnesses = reader.read_serializable_list(verification.Witness, max=1)
        if len(witnesses) != 1:
            raise ValueError(
                f"Deserialization error - Witness object count is {len(witnesses)} must be 1"
            )
        self.witness = witnesses[0]

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the unsigned data part of the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the primary_index field is greater than the configured consensus validator count.
        """
        self.version = reader.read_uint32()
        if self.version > 0:
            raise ValueError("Deserialization error - invalid version")
        self.prev_hash = reader.read_serializable(types.UInt256)
        self.merkle_root = reader.read_serializable(types.UInt256)
        self.timestamp = reader.read_uint64()
        self.nonce = reader.read_uint64()
        self.index = reader.read_uint32()
        self.primary_index = reader.read_uint8()
        self.next_consensus = reader.read_serializable(types.UInt160)

    def get_script_hashes_for_verifying(self, snapshot) -> list[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        if self.prev_hash == types.UInt256.zero():
            return [self.witness.script_hash()]
        prev_block = snapshot.blocks.try_get(self.prev_hash, read_only=True)
        if prev_block is None:
            raise ValueError(
                "Can't get next_consensus hash from previous block. Block does not exist"
            )
        return [prev_block.next_consensus]

    @classmethod
    def _serializable_init(cls):
        return cls(
            0,
            types.UInt256.zero(),
            0,
            0,
            0,
            0,
            types.UInt160.zero(),
            verification.Witness(b"", b""),
        )


class Block(inventory.IInventory):
    """
    The famous Block. I transfer chain state.
    """

    def __init__(
        self,
        header: Header,
        transactions: Optional[Sequence[transaction.Transaction]] = None,
        *args,
        **kwargs,
    ):
        super(Block, self).__init__(*args, **kwargs)
        self.header = header
        self.transactions = [] if transactions is None else transactions

    def __len__(self):
        # calculate the varint length that needs to be inserted before the transaction objects.
        magic_len = utils.get_var_size(len(self.transactions))
        txs_len = sum([len(t) for t in self.transactions])
        return len(self.header) + magic_len + txs_len

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if self.hash() != other.hash():
            return False
        return True

    @property
    def version(self) -> int:
        """Block data structure version - for internal use."""
        return self.header.version

    @property
    def prev_hash(self) -> types.UInt256:
        """The hash of the previous block."""
        return self.header.prev_hash

    @property
    def merkle_root(self) -> types.UInt256:
        """The merkle root of the transactions in the block."""
        return self.header.merkle_root

    @property
    def timestamp(self) -> int:
        """UTC timestamp in milliseconds."""
        return self.header.timestamp

    @property
    def nonce(self) -> int:
        """Random number."""
        return self.header.nonce

    @property
    def index(self) -> int:
        """The height of the block."""
        return self.header.index

    @property
    def primary_index(self) -> int:
        """The index into the consensus node list that was used to generate this block."""
        return self.header.primary_index

    @property
    def next_consensus(self) -> types.UInt160:
        """The hash of the consensus node that will generate the next block."""
        return self.header.next_consensus

    @property
    def witness(self) -> verification.Witness:
        """The witness of this block."""
        return self.header.witness

    def hash(self) -> types.UInt256:
        """A unique identifier based on the unsigned data portion of the object."""
        return self.header.hash()

    def get_script_hashes_for_verifying(self, snapshot) -> list[types.UInt160]:
        """
        Helper method to get the data used in verifying the object.
        """
        return self.header.get_script_hashes_for_verifying(snapshot)

    @property
    def inventory_type(self) -> inventory.InventoryType:
        """
        Inventory type identifier.
        """
        return inventory.InventoryType.BLOCK

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_var_int(len(self.transactions))
        for tx in self.transactions:
            writer.write_serializable(tx)

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the unsigned part of the object into a binary stream.

        Args:
            writer: instance.
        """
        self.header.serialize_unsigned(writer)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if the content count of the block is zero, or if there is a duplicate transaction in the list,
                or if the merkle root does not include the calculated root.
        """
        self.header = reader.read_serializable(Header)
        self.transactions = reader.read_serializable_list(
            transaction.Transaction, max=0xFFFF
        )

        if len(set(self.transactions)) != len(self.transactions):
            raise ValueError(
                "Deserialization error - block contains duplicate transaction"
            )

        hashes = [t.hash() for t in self.transactions]
        if crypto.MerkleTree.compute_root(hashes) != self.header.merkle_root:
            raise ValueError("Deserialization error - merkle root mismatch")

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        """Not supported."""
        raise NotImplementedError

    def rebuild_merkle_root(self) -> None:
        """
        Recalculates the Merkle root.
        """
        self.header.merkle_root = crypto.MerkleTree.compute_root(
            [t.hash() for t in self.transactions]
        )

    def trim(self) -> TrimmedBlock:
        """
        Reduce a block in size by replacing the transaction objects with their identifying hashes.
        """
        return TrimmedBlock(self.header, [t.hash() for t in self.transactions])

    @classmethod
    def _serializable_init(cls):
        return cls(Header._serializable_init(), [])


class TrimmedBlock(serialization.ISerializable):
    """
    A size reduced `Block` instance.

    Contains consensus data and transactions hashes instead of their full objects.
    """

    def __init__(self, header: Header, hashes: Sequence[types.UInt256]):
        super(TrimmedBlock, self).__init__()
        self.header = header
        self.hashes = hashes

    def __len__(self):
        return len(self.header) + utils.get_var_size(self.hashes)

    def __deepcopy__(self, memodict={}):
        # not the best, but faster than letting deepcopy() do introspection
        return self.__class__.deserialize_from_bytes(self.to_array())

    def hash(self):
        """A unique identifier based on the unsigned data portion of the object."""
        return self.header.hash()

    @property
    def index(self):
        """The height of the block."""
        return self.header.index

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_serializable_list(list(self.hashes))

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.header = reader.read_serializable(Header)
        self.hashes = reader.read_serializable_list(types.UInt256, max=0xFFFF)

    @classmethod
    def _serializable_init(cls):
        return cls(Header._serializable_init(), [])


class MerkleBlockPayload(serialization.ISerializable):
    """
    Payload for transfering merkletree hashes of a block.
    """

    def __init__(self, block: Block, flags: bitarray):
        hashes = [t.hash() for t in block.transactions]
        tree = crypto.MerkleTree(hashes)
        self.flags = flags.tobytes()
        self.tx_count = len(hashes)
        self.hashes = tree.to_hash_array()
        self.header = block.header

    def __len__(self):
        return (
            len(self.header)
            + s.uint32
            + utils.get_var_size(self.hashes)
            + utils.get_var_size(self.flags)
        )

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.header)
        writer.write_var_int(self.tx_count)
        writer.write_serializable_list(self.hashes)
        writer.write_var_bytes(self.flags)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.header = reader.read_serializable(Header)
        self.tx_count = reader.read_var_int(max=0xFFFF)
        self.hashes = reader.read_serializable_list(types.UInt256, max=self.tx_count)
        self.flags = reader.read_var_bytes(max=(max(self.tx_count, 1) + 7) // 8)

    @classmethod
    def _serializable_init(cls):
        block = Block._serializable_init()
        flags = bitarray()
        return cls(block, flags)


class HeadersPayload(serialization.ISerializable):
    """
    Payload for sending or receiving `Block` headers.
    """

    MAX_HEADERS_COUNT = 2000

    def __init__(self, headers: Optional[Sequence[Header]] = None):
        self.headers = list(headers) if headers else []

    def __len__(self):
        return utils.get_var_size(self.headers)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable_list(self.headers)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.headers = reader.read_serializable_list(Header)


class GetBlocksPayload(serialization.ISerializable):
    """
    Used to request an array of block hashes that can be retrieved via a message with the
    `message.MessageType.GETDATA` type.
    """

    def __init__(self, hash_start: types.UInt256, count=-1):
        """
        Create payload.

        Args:
            hash_start: starting point from which to return the `next` hash.

                Note:

                    For syncing supply the local best height block hash to receive the hashes in the range of
                    `best_height`+`1` to `best_height`+`1`+`count`.

            count: number of hashes to return.
        """
        self.hash_start = hash_start
        self.count = count

    def __len__(self):
        return s.uint256 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.hash_start)
        writer.write_int16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.hash_start = reader.read_serializable(types.UInt256)
        self.count = reader.read_int16()

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt256.zero())


class GetBlockByIndexPayload(serialization.ISerializable):
    """
    Used to request full Block or Header objects via a message with the `message.MessageType.GETBLOCKBYINDEX`
     or `message.MessageType.GETHEADERS` type respectively.
    """

    def __init__(self, index_start: int, count: int = HeadersPayload.MAX_HEADERS_COUNT):
        """
        Create payload.

        Args:
            index_start: start block height.
            count: number of blocks or headers to requests starting from `index_start`.

        """
        self.index_start = index_start
        self.count = count

    def __len__(self):
        return s.uint32 + s.uint16

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.index_start)
        writer.write_int16(self.count)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.

        Raises:
            ValueError: if `count` is zero or exceeds `GetBlockByIndexPayload.MAX_BLOCKS_COUNT`.
        """
        self.index_start = reader.read_uint32()
        self.count = reader.read_int16()
        if (
            self.count < 1
            or self.count == 0
            or self.count > HeadersPayload.MAX_HEADERS_COUNT
        ):
            raise ValueError("Deserialization error - invalid count")

    @classmethod
    def _serializable_init(cls):
        return cls(0)



"""
Neo Executable Format.
"""

from __future__ import annotations
import base64
import hashlib
from collections.abc import Sequence
from neo3.core import serialization, types, Size as s, utils as coreutils, interfaces
from neo3.contracts import callflags
from typing import Optional


class NEF(serialization.ISerializable, interfaces.IJson):
    """
    Neo Executable Format container.
    """

    def __init__(
        self,
        compiler_name: Optional[str] = None,
        script: Optional[bytes] = None,
        tokens: Optional[Sequence[MethodToken]] = None,
        source: Optional[str] = None,
        _magic: int = 0x3346454E,
    ):
        """
        Args:
            compiler_name: human-readable name of the compiler and version used to create the script data.
            Automatically limited to 64 bytes
            script: a byte array of raw VM opcodes.
            tokens:
            source: url where source files can be found
        """
        self.magic = _magic
        if compiler_name is None:
            self.compiler = "unknown"
        else:
            self.compiler = compiler_name[:64]
        self.source = source if source else ""
        self.script = script if script else b""
        self._checksum = 0
        self.tokens = [] if tokens is None else list(tokens)
        # this is intentional, because NEO computes the initial checksum by serializing itself while checksum is 0
        self._checksum = self.compute_checksum()

    def __len__(self):
        return (
            s.uint32  # magic
            + 64  # compiler
            + coreutils.get_var_size(self.source)
            + 1  # reserved
            + coreutils.get_var_size(self.tokens)
            + 2  # reserved
            + s.uint32  # checksum
            + coreutils.get_var_size(self.script)
        )

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            self.magic == other.magic
            and self.compiler == other.compiler
            and self.source == other.source
            and self.script == other.script
            and self.tokens == other.tokens
            and self.checksum == other.checksum
        )

    @property
    def checksum(self) -> int:
        """
        Data integrity value.
        """
        if self._checksum == 0:
            self._checksum = self.compute_checksum()
        return self._checksum

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.magic)
        writer.write_bytes(self.compiler.encode("utf-8").ljust(64, b"\x00"))
        writer.write_var_string(self.source)
        writer.write_bytes(b"\x00")
        writer.write_serializable_list(self.tokens)
        writer.write_bytes(b"\x00\x00")
        writer.write_var_bytes(self.script)
        writer.write_uint32(self._checksum)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        if reader.read_uint32() != self.magic:
            raise ValueError("Deserialization error - Incorrect magic")
        self.compiler = reader.read_bytes(64).decode("utf-8").rstrip(b"\x00".decode())
        self.source = reader.read_var_string(256)
        if reader.read_uint8() != 0:
            raise ValueError("Reserved bytes must be 0")
        self.tokens = reader.read_serializable_list(MethodToken)
        if reader.read_uint16() != 0:
            raise ValueError("Reserved bytes must be 0")

        self.script = reader.read_var_bytes(max=0xFFFF * 2)
        if len(self.script) == 0:
            raise ValueError("Deserialization error - Script can't be empty")

        checksum = int.from_bytes(reader.read_bytes(4), "little")
        if checksum != self.compute_checksum():
            raise ValueError("Deserialization error - Invalid checksum")
        else:
            self._checksum = checksum

    def compute_checksum(self) -> int:
        """
        Compute the checksum of the NEF file.
        """
        return int.from_bytes(
            hashlib.sha256(hashlib.sha256(self.to_array()[:-4]).digest()).digest()[:4],
            "little",
        )

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        raise NotImplementedError

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON."""
        tokens = []
        for t in json["tokens"]:
            tokens.append(MethodToken.from_json(t))
        script = base64.b64decode(json["script"])
        return cls(json["compiler"], script, tokens, _magic=json["magic"])

    @classmethod
    def from_file(cls, path: str):
        """
        Create object from a file.

        Args:
            path: location of the file.

        Raises:
            FileNotFoundError: if the path is invalid.
            ValueError: if the file is not a valid NEF.
        """
        with open(path, "rb") as f:
            try:
                return cls.deserialize_from_bytes(f.read())
            except ValueError as e:
                raise ValueError(f"Failed NEF file validation with: {e}")

    @classmethod
    def _serializable_init(cls):
        c = cls()
        c._checksum = 0
        return c


class MethodToken(serialization.ISerializable, interfaces.IJson):
    def __init__(
        self,
        hash: types.UInt160,
        method: str,
        parameters_count: int,
        has_return_value: bool,
        call_flags: callflags.CallFlags,
    ):
        self.hash = hash
        self.method = method
        self.parameters_count = parameters_count
        self.has_return_value = has_return_value
        self.call_flags = call_flags

    def __len__(self):
        return (
            s.uint160
            + coreutils.get_var_size(self.method)
            + s.uint16
            + s.uint8
            + s.uint8
        )

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_serializable(self.hash)
        writer.write_var_string(self.method)
        writer.write_uint16(self.parameters_count)
        writer.write_uint8(self.has_return_value)
        writer.write_uint8(self.call_flags.value)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.hash = reader.read_serializable(types.UInt160)
        self.method = reader.read_var_string(32)
        self.parameters_count = reader.read_uint16()
        self.has_return_value = bool(reader.read_uint8())
        self.call_flags = callflags.CallFlags(reader.read_uint8())

    def to_json(self) -> dict:
        """Convert object into JSON representation."""
        raise NotImplementedError

    @classmethod
    def from_json(cls, json: dict):
        """Create object from JSON"""
        h = types.UInt160.from_string(json["hash"][2:])
        m = json["method"]
        pc = json["paramcount"]
        rv = json["hasreturnvalue"]
        cf = callflags.CallFlags.from_csharp_name(json["callflags"])
        return cls(h, m, pc, rv, cf)

    @classmethod
    def _serializable_init(cls):
        return cls(types.UInt160.zero(), "", 0, False, callflags.CallFlags.NONE)
